import sys
import os
import shutil
import contextlib
from pathlib import Path
from typing import Collection

import click
import ansible_runner

from micado.installer.ansible.playbook import Playbook

DEFAULT_VERS = "v0.11.2"

CONFIGS: dict = {
    "hosts": ("playbook/inventory", "hosts.yml"),
    "cloud": ("playbook/project/credentials", "credentials-cloud-api.yml"),
    "registry": ("playbook/project/credentials", "credentials-docker-registry.yml"),
    "micado": ("playbook/project/credentials", "credentials-micado.yml"),
    "settings": ("playbook/project/host_vars", "micado.yml"),
}


class OrderedGroup(click.Group):
    def list_commands(self, ctx) -> Collection:
        return self.commands.keys()


@click.group(cls=OrderedGroup)
@click.pass_context
def cli(ctx):
    """The MiCADO Command Line Interface.

    \b
    A typical workflow consists of:
      `micado init`    To gather setup files
      `micado config`  To configure the deployment
      `micado deploy`  To install MiCADO
    """
    if ctx.invoked_subcommand == "init":
        pass
    elif not Path(".micado").absolute().is_dir():
        click.secho("The current directory is not initialised. ", fg="yellow", nl=False)
        click.secho(
            "Please initalise a new directory with `micado init`. ", fg="yellow"
        )
        sys.exit(1)


@cli.command()
@click.argument(
    "target",
    required=False,
    default=".",
    type=click.Path(file_okay=False, writable=True, readable=True, resolve_path=True),
)
@click.option(
    "--version",
    required=False,
    default=DEFAULT_VERS,
    type=str,
    help="MiCADO semantic version, prefixed by v (e.g. v0.11.2)",
)
@click.option(
    "--force",
    is_flag=True,
    help="""Reset the MiCADO setup directory. 
    WARNING: This will reset any MiCADO settings from this directory.""",
)
def init(target, version, force):
    """Initalises a TARGET directory with VERSION setup files

    Uses current directory if no TARGET, current MiCADO if no VERSION
    """
    target_text = click.style(f"{target}", italic=True, fg="reset")
    if force:
        reset_directory(target)

    elif directory_is_not_empty(target):
        click.secho(f"The directory {target_text}", fg="yellow", nl=False)
        click.secho(" is not empty. Is MiCADO already initialised?", fg="yellow")
        sys.exit(1)

    os.makedirs(target, exist_ok=True)

    playbook = Playbook(version, f"{os.getlogin()}-cli", f"{target}/")
    playbook.playbook_path = f"{target}/.micado"
    try:
        playbook.download()
    except TypeError:
        click.secho(f"Cannot find MiCADO version {version}", fg="red")
        sys.exit(1)
    playbook.extract()

    click.secho(
        f"Succesfully initialised the MiCADO setup in {target_text}", fg="green"
    )


@cli.group()
def config():
    """Configure details of a MiCADO cluster before deployment."""
    pass


@config.command()
def hosts():
    """Specify connection details for the MiCADO host (REQUIRED)"""
    open_config_file("hosts")


@config.command()
def cloud():
    """Specify cloud API credentials (REQUIRED)"""
    open_config_file("cloud")


@config.command()
def registry():
    """Specify private Docker registry credentials"""
    open_config_file("registry")


@config.command()
def micado():
    """Configure MiCADO WebUI access settings"""
    open_config_file("micado")


@config.command()
def settings():
    """Configure common deployment settings"""
    open_config_file("settings")


@cli.command()
@click.option(
    "--vault",
    is_flag=True,
    help="Asks for the vault password. (Required if using vault)",
)
def deploy(vault):
    """Deploys a MiCADO cluster as per the configuration"""
    if not os.path.exists(CONFIGS["hosts"][1]):
        click.secho(f"MiCADO host not configured! Use `micado config hosts`", fg="red")
        sys.exit(1)
    if not os.path.exists(CONFIGS["cloud"][1]):
        click.secho(
            f"Target clouds not configured! Use `micado config cloud`", fg="red"
        )
        sys.exit(1)

    password = (
        click.prompt("Enter the vault password", type=str, hide_input=True)
        if vault
        else ""
    )
    cmdline = "--ask-vault-pass" if vault else " "
    passwords = {"^Vault password:\\s*?$": password} if vault else {}

    ansible_runner.run(
        playbook="micado.yml",
        cmdline=cmdline,
        passwords=passwords,
        private_data_dir="./.micado/playbook",
    )


def directory_is_not_empty(dir) -> bool:
    try:
        return bool(os.listdir(dir))
    except FileNotFoundError:
        return False


def remove_sample_from_filename(file):
    path = Path(".micado").absolute() / Path(file[0]) / Path(file[1])
    try:
        src = path.parent / f"sample-{path.name}"
        dst = path.parent / f"{path.name}"
        shutil.move(src, dst)
    except FileNotFoundError:
        pass


def get_symlink_config_file(file) -> str:
    path = Path(".micado").absolute() / Path(file[0]) / Path(file[1])
    try:
        Path(path.name).absolute().symlink_to(path)
    except FileNotFoundError:
        raise
    except FileExistsError:
        pass
    return str(Path(path.name).absolute())


def produce_credential_warning(file):
    click.secho(
        "Please consider encrypting credential files with ansible-vault.", bold=True
    )
    click.echo("Use the same vault password across all files in this setup.\n")
    click.echo(
        "If you have ansible-vault installed, you may use the following command:"
    )
    click.secho(f"    ansible-vault encrypt {file[1]}\n", italic=True)
    click.echo("If you need to edit the file again, first decrypt it with:")
    click.secho(f"    ansible-vault decrypt {file[1]}", italic=True)


def reset_directory(target):
    with contextlib.suppress(FileNotFoundError):
        shutil.rmtree(Path(target) / ".micado")
        for file in CONFIGS.values():
            (Path(target) / file[1]).unlink()


def open_config_file(choice):
    file: tuple[str, str] = CONFIGS[choice]
    remove_sample_from_filename(file)
    try:
        symlink = get_symlink_config_file(file)
    except FileNotFoundError:
        click.secho("Could not find the config file.", fg="red")
        click.secho("  Reset all files with `micado init . --force`")
        sys.exit(1)

    try:
        click.edit(filename=symlink)
    except click.UsageError:
        click.secho("Could not open default text editor.", fg="red")
        click.secho("  You can manually edit the file at {symlink}.")

    click.secho(f"Symlink created at: {symlink}\n", fg="green")
    if file[0].endswith("credentials"):
        produce_credential_warning(file)



"""

Low-level MiCADO API client, to be used by the higher-level ..client

"""

import requests

from .application import ApplicationMixin


class SubmitterClient(requests.Session, ApplicationMixin):
    """Low-level MiCADO client, prefer use of `micado.client`

    Args:
        endpoint (string): The endpoint of the running submitter.
        version (string, optional): API version. Defaults to 'v2.0'.
        verify (bool or string, optional): Whether to verify the
            connection with certificate on the client side. For
            self-signed certificates, this parameter can be a string
            path to the .pem certificate. Defaults to True.
        auth (tuple, optional): Basic auth username and password.
            Defaults to None.

    Raises:
        TypeError: If auth is in poorly formatted

    """

    def __init__(self, endpoint, version="v2.0", verify=True, auth=None):
        super().__init__()
        self.endpoint = endpoint.strip("/") + "/"
        self._version = version
        self.verify = verify
        if isinstance(auth, tuple):
            self.auth = auth
        elif auth:
            raise TypeError("Basic auth must be a tuple of (<user>, <pass>)")

    def _url(self, path):
        return self.endpoint + self._version + path



from requests.exceptions import HTTPError


class MicadoException(Exception):
    """Base class for exceptions"""

class MicadoAPIException(MicadoException):
    """Class for API exceptions"""

def detailed_raise_for_status(resp):
    """Invoke raise_for_status on a Response object and add message

    Args:
        resp (Response): requests.Response object

    Raises:
        Exception: MicadoException or the original HTTPError
    """
    try:
        resp.raise_for_status()
    except HTTPError as e:
        if resp.text:
            raise MicadoAPIException(
                resp.json().get("message", resp.text)
            ) from None
        else:
            raise e



import pytest
import requests
from unittest.mock import Mock, patch

from micado.launcher.openstack import auth

AUTH_CONSTRUCTOR_CALLS = [
    (
        auth.PasswordAuthenticator,
        {
            "username": "user",
            "password": "pass",
        },
    ),
    (
        auth.AppCredAuthenticator,
        {
            "application_credential_id": "aci",
            "application_credential_secret": "acs",
        },
    ),
    (
        auth.OidcAuthenticator,
        {
            "identity_provider": "ip",
            "access_token": "at",
        },
    ),
]


def test_abstract_authenticator():
    with pytest.raises(NotImplementedError):
        auth.Authenticator().authenticate()


@pytest.mark.parametrize("authenticator,args", AUTH_CONSTRUCTOR_CALLS)
def test_concrete_auth_correct_args(authenticator, args):
    authenticator(**args)


@pytest.mark.parametrize("authenticator,args", AUTH_CONSTRUCTOR_CALLS)
def test_password_auth_missing_args(authenticator, args):
    args.pop(list(args.keys())[0])
    with pytest.raises(TypeError):
        authenticator(**args)


@pytest.mark.parametrize("authenticator,args", AUTH_CONSTRUCTOR_CALLS)
def test_password_auth_too_many_args(authenticator, args):
    args["extra"] = "extra"
    with pytest.raises(TypeError):
        authenticator(**args)

def test_oidc_calls_openid_refresh_with_dict(monkeypatch):
    mocked_fn = Mock()
    monkeypatch.setattr(auth, "refresh_openid_token", mocked_fn)
    auth.OidcAuthenticator(access_token={}, identity_provider="")
    assert(mocked_fn.called)


@pytest.fixture
def requests_patch(monkeypatch, request):
    def patched_post(*args, **kwargs):
        mocked_response = Mock()
        mocked_response.json = lambda: {request.param: kwargs}
        return mocked_response

    monkeypatch.setattr(requests, "post", patched_post)


@pytest.mark.parametrize("requests_patch", ["access_token"], indirect=True)
def test_openid_token_refresh_request_body(requests_patch):
    url = "http://test.com"
    client_id, client_secret, token = "cli_id", "cli_sec", "ref_tok"
    data = {
        "client_id": client_id,
        "client_secret": client_secret,
        "grant_type": "refresh_token",
        "refresh_token": token,
        "scope": "openid email profile",
    }
    response = auth.refresh_openid_token(
        url=url,
        client_id=client_id,
        client_secret=client_secret,
        refresh_token=token,
    )
    assert response["url"] == url
    assert response["auth"] == (client_id, client_secret)
    assert response["data"] == data



"""Module for managing and representing applications in MiCADO

"""

from .base import Model, Resource


class Application(Model):
    """Representation of an application deployed in MiCADO

    The state of an application can be refreshed with reload()
    """

    @property
    def adaptors(self):
        """
        Adaptor info for this application
        """
        return self.info.get("adaptors")


class Applications(Resource):
    """Model for managing applications in MiCADO

    Basic CRUD functionality is implemented here with
    create(), list()/get(), update() and delete()

    """

    model = Application

    def __init__(self, client):
        self.client = client

    def get(self, app_id):
        """Retrieves info on a specific application, given its ID

        Args:
            app_id (string): Application ID to fetch. Required

        Usage:

            >>> my_app = client.applications.get("stresstest")
            >>> my_app.id
            "stresstest"
            >>> my_app.adaptors
            {'KubernetesAdaptor': 'Executed', 'OccopusAdaptor': 'Skipped'}

        Returns:
            Application object: Relevant information for a single application
        """
        return self._make_model(app_id, self.client.api.inspect_app(app_id))

    def list(self):
        """Retrieves the available list of applications in MiCADO

        Usage:

            >>> running_apps = client.applications.list()
            >>> [app.id for app in running_apps]
            ["stresstest"]

        Returns:
            list of Application objects: Relevant info for all applications
        """
        app_ids = self.client.api.applications()
        return [self.get(i) for i in app_ids]

    def create(self, app_id=None, **kwargs):
        """Creates a new application in MiCADO

        Args:
            app_id (string, optional): Application ID. Generated if None.
            adt (dict, optional): YAML dict of Application Description
                Template. Required if URL is empty. Defaults to None.
            url (string, optional): URL of YAML ADT. Required if ADT is empty.
                Defaults to None.
            params (dict, optional): TOSCA input parameters. Defaults to {}.
            dryrun (bool, optional): Flag to skip execution of components.
                Defaults to False.

        Usage:

            >>> client.applications.create(app_id="stresstest",
                                           url="example.com/repo/adt.yaml")
            "stresstest created successfully"

        Returns:
            dict: ID and status of deployment
        """
        kwargs["app_id"] = app_id
        return self.client.api.create_app(**kwargs)

    def delete(self, app_id):
        """Deletes an application in MiCADO given its ID.

        Args:
            app_id (string): Application ID to delete

        Usage:

            >>> client.applications.delete("stresstest")
            "stresstest deleted successfully"

        Returns:
            dict: ID and status of deletion
        """
        return self.client.api.delete_app(app_id)



"""

Higher-level client for both submitter interaction and launching ability

"""

import os

from .api.client import SubmitterClient

from .launcher.openstack import OpenStackLauncher
from .launcher.cloudbroker import CloudBrokerLauncher

from .installer.ansible import AnsibleInstaller

from .models.application import Applications
from .models.micado import Micado

from .exceptions import MicadoException

LAUNCHERS = {
    "openstack": OpenStackLauncher,
    "cloudbroker": CloudBrokerLauncher,
}

INSTALLER = {
    "ansible": AnsibleInstaller,
}


class MicadoClient:
    """The MiCADO Client

    Builds and communicates with a MiCADO node

    Usage with a launcher:

    a)

        >>> from micado import MicadoClient
        >>> client = MicadoClient(launcher="openstack", installer="ansible")
        >>> client.micado.create(
        ...     auth_url='yourendpoint',
        ...     project_id='project_id',
        ...     image='image_name or image_id',
        ...     flavor='flavor_name or flavor_id',
        ...     network='network_name or network_id',
        ...     keypair='keypair_name or keypair_id',
        ...     security_group='security_group_name or security_group_id'
        ... )
        >>> client.applications.list()
        >>> client.micado.destroy()

    b)

        >>> from micado import MicadoClient
        >>> client = MicadoClient(launcher="openstack", installer="ansible")
        >>> micado_id = client.micado.create(
        ...     auth_url='yourendpoint',
        ...     project_id='project_id',
        ...     image='image_name or image_id',
        ...     flavor='flavor_name or flavor_id',
        ...     network='network_name or network_id',
        ...     keypair='keypair_name or keypair_id',
        ...     security_group='security_group_name or security_group_id'
        ... )
        >>> client.applications.list()
        >>> << store your micado_id >>
        >>> << exit >>
        >>> -------------------------------------------------------------
        >>> << start >>
        >>> ...
        >>> micado_id = << retrieve micado_id >>
        >>> client = MicadoClient(launcher="openstack", installer="ansible")
        >>> client.micado.attach(micado_id = micado_id)
        >>> client.applications.list()
        >>> client.micado.destroy()

    c)

        >>> from micado import MicadoClient
        >>> client = MicadoClient(launcher="cloudbroker", installer="ansible")
        >>> client.micado.create(
        ...     auth_url='yourendpoint',
        ...     deployment_id='deployment_id',
        ...     instance_type_id='image_name or image_id',
        ...     key_pair_id='key_pair_id',
        ...     firewall_rule_set_id='firewall_rule_set_id',
        ... )
        >>> client.applications.list()
        >>> client.micado.destroy()


    Usage without a launcher i.e. MiCADO is already created independently from the client library.

        >>> from micado import MicadoClient
        >>> client = MicadoClient(
        ...     endpoint="https://micado/toscasubmitter/",
        ...     version="v2.0",
        ...     verify=False,
        ...     auth=("ssl_user", "ssl_pass")
        ... )
        >>> client.applications.list()

    Args:
        auth_url (string): Authentication URL for the NOVA
            resource.
        image (string): Name or ID of the image resource.
        flavor (string): Name or ID of the flavor resource.
        network (string): Name or ID of the network resource.
        keypair (string): Name or ID of the keypair resource.
        security_group (string, optional): name or ID of the
            security_group resource. Defaults to 'all'.
        region (string, optional): Name of the region resource.
            Defaults to None.
        user_domain_name (string, optional): Define the user_domain_name.
            Defaults to 'Default'
        project_id (string, optional): ID of the project resource.
            Defaults to None.
        micado_user (string, optional): MiCADO username.
            Defaults to admin.
        micado_password (string, optional): MiCADO password.
            Defaults to admin.
        enable_terraform (bool, optional): Enable terraform in the deployment.
            Defaults to False.
        endpoint (string): Full URL to API endpoint (omit version).
            Required.
        version (string, optional): MiCADO API Version (minimum v2.0).
            Defaults to 'v2.0'.
        verify (bool, optional): Verify certificate on the client-side.
            OR (str): Path to cert bundle (.pem) to verfiy against.
            Defaults to True.
        auth (tuple, optional): Basic auth credentials (<user>, <pass>).
            Defaults to None.
    """

    def __init__(self, *args, **kwargs):
        launcher = kwargs.pop("launcher", "").lower()
        if launcher:
            self.api = None
            try:
                self.launcher = LAUNCHERS[launcher]()
            except KeyError:
                raise MicadoException(f"Unknown launcher: {launcher}")
            try:
                installer = kwargs.pop("installer", "").lower()
                if installer != '':
                    self.installer = INSTALLER[installer]()
            except KeyError:
                raise MicadoException(f"Unknown installer: {installer}")
        else:
            self.api = SubmitterClient(*args, **kwargs)

    @classmethod
    def from_env(cls):
        """Usage:
            Ensure MICADO_API_ENDPOINT and MICADO_API_VERSION
            environment variables are set, then:

            >>> from micado import MicadoClient
            >>> client = MicadoClient.from_env()
        """
        try:
            submitter_endpoint = os.environ["MICADO_API_ENDPOINT"]
            submitter_version = os.environ["MICADO_API_VERSION"]
        except KeyError as err:
            raise MicadoException(f"Environment variable {err} not defined!")

        return cls(
            endpoint=submitter_endpoint,
            version=submitter_version,
            verify=False,
        )

    @property
    def applications(self):
        return Applications(client=self)

    @property
    def micado(self):
        if not self.launcher:
            raise MicadoException("No launcher defined")
        return Micado(client=self)



"""
Higher-level methods to manage the MiCADO node
"""
import os
from pathlib import Path

from micado.utils.utils import DataHandling

from ..api.client import SubmitterClient
from .base import Model

DEFAULT_PATH = Path.home() / ".micado-cli"


class Micado(Model):
    home = str(Path(os.environ.get("MICADO_CLI_DIR", DEFAULT_PATH))) + "/"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def micado_id(self):
        return self.client.micado_id

    @micado_id.setter
    def micado_id(self, micado_id):
        self.client.micado_id = micado_id

    @property
    def micado_ip(self):
        return self.client.micado_ip

    @micado_ip.setter
    def micado_ip(self, micado_ip):
        self.client.micado_ip = micado_ip

    @property
    def details(self):
        return self.client.details
    
    @details.setter
    def details(self, details):        
        self.client.details = details

    @property
    def launcher(self):
        return self.client.launcher

    @property
    def installer(self):
        return self.client.installer

    @property
    def api(self):
        return self.client.api

    @api.setter
    def api(self, api):
        self.client.api = api

    def init_api(self):
        """Configure Submitter API

        Returns:
            SubmitterClient: return SubmitterClient
        """
        server = DataHandling.get_properties(f"{self.home}data.yml", self.micado_id)
        self.micado_ip = server["ip"]
        
        self.details = f"""
        WebUI: https://{server["ip"]}
        Username: {server["micado_user"]}
        Password: {server["micado_password"]}
        """
        
        return SubmitterClient(
            endpoint=server["endpoint"],
            version=server["api_version"],
            verify=server["cert_path"],
            auth=(server["micado_user"], server["micado_password"]),
        )

    def attach(self, micado_id):
        """Configure the micado object to handle the instance
        created by the def:create()

        Args:
            micado_id (string): micado ID returned by def:create()
        """
        self.micado_id = micado_id
        self.api = self.init_api()

    def create(self, **kwargs):
        """Creates a new MiCADO VM and deploy MiCADO services on it.

        Args:
            auth_url (string): Authentication URL for the NOVA
                resource.
            image (string): Name or ID of the image resource.
            flavor (string): Name or ID of the flavor resource.
            network (string): Name or ID of the network resource.
            keypair (string): Name or ID of the keypair resource.
            security_group (string, optional): name or ID of the
                security_group resource. Defaults to 'all'.
            region (string, optional): Name of the region resource.
                Defaults to None.
            user_domain_name (string, optional): Define the user_domain_name.
                Defaults to 'Default'
            project_id (string, optional): ID of the project resource.
                Defaults to None.
            micado_user (string, optional): MiCADO username.
                Defaults to admin.
            micado_password (string, optional): MiCADO password.
                Defaults to admin.

        Usage:

            >>> client.micado.create(
            ...     auth_url='yourendpoint',
            ...     project_id='project_id',
            ...     image='image_name or image_id',
            ...     flavor='flavor_name or flavor_id',
            ...     network='network_name or network_id',
            ...     keypair='keypair_name or keypair_id',
            ...     security_group='security_group_name or security_group_id'
            ... )

        Returns:
            string: ID of MiCADO

        """
        try:
            _micado = self.launcher.launch(**kwargs)
            self.micado_id = _micado.id
            self.micado_ip = _micado.ip

            self.installer.deploy(_micado, **kwargs)
            self.api = self.init_api()
        except Exception as e:
            if hasattr(locals()["self"], "micado_id"):
                self.launcher.delete(self.micado_id)
            raise
        return self.micado_id

    def destroy(self):
        """Destroy running applications and the existing MiCADO VM.

        Usage:

            >>> client.micado.destroy()

        """
        self.api = self.init_api()
        self.api._destroy()
        self.api = None
        self.launcher.delete(self.micado_id)



import os
import shutil
import tarfile
from pathlib import Path

import ansible_runner
import requests

PLAYBOOK_NAME = "micado.yml"
PLAYBOOK_INTERNAL = "playbook"
ROTATION = 100  # max number of artifacts (logs, etc...) to keep
QUIET = True  # hide ansible output


class Playbook:
    def __init__(self, version: str, id: str, home_dir: str):
        self.version: str = version
        self.id: str = id
        self.tar_download: Path = Path(f"{home_dir}micado-{version}.tar.gz")
        self.playbook_path: Path = Path(f"{home_dir}micado-{version}")

    def run(self, hosts: dict, extravars: dict, playbook: str = None):
        """Run the playbook"""
        if not self.playbook_exists():
            self.download()
            self.extract()
        
        data_dir = self.playbook_path / PLAYBOOK_INTERNAL

        # fix_hosts_permissions() because https://github.com/ansible/ansible-runner/issues/853
        fix_hosts_permissions(data_dir)
        runner = ansible_runner.interface.run(
            ident=self.id,
            playbook=playbook or PLAYBOOK_NAME,
            private_data_dir=str(data_dir),
            inventory=hosts,
            extravars=extravars,
            rotate_artifacts=ROTATION,
            quiet=QUIET,
        )
        fix_hosts_permissions(data_dir)
        return runner

    def download(self):
        """Download playbook from GitHub and write down to home directory."""
        url = f"https://github.com/micado-scale/ansible-micado/tarball/{self.version}"
        r = requests.get(url, stream=True)

        with open(self.tar_download, "wb") as f:
            f.write(r.content)

        if not tarfile.is_tarfile(self.tar_download):
            os.remove(self.tar_download)
            raise TypeError(f"Download failed - check MiCADO {self.version} exists.")

    def extract(self):
        """Extract tar to the directory where it was downloaded"""
        if not os.path.isfile(self.tar_download):
            raise FileNotFoundError("Playbook tarball not found. Cannot extract.")

        tar_file = tarfile.open(self.tar_download)
        tar_path = self.tar_download.parent / tar_file.firstmember.name
        tar_file.extractall(self.tar_download.parent)
        tar_file.close()

        # Don't overwrite an existing playbook of the same version
        try:
            tar_path.rename(self.playbook_path)
        except OSError:
            shutil.rmtree(str(tar_path))

        self.tar_download.unlink()  # delete the tarball

    def remove(self):
        """Remove the playbook directory"""
        if not self.playbook_path:
            raise FileNotFoundError("Playbook directory not found. Cannot remove.")
        shutil.rmtree(str(self.playbook_path))
        self.playbook_path = None

    def playbook_exists(self):
        """Check if playbook directory exists"""
        return os.path.isdir(self.playbook_path)

def fix_hosts_permissions(path: Path):
    try:
        os.chmod(path / "inventory/hosts.json", 0o600)
    except FileNotFoundError:
        pass


import requests
from keystoneauth1.identity import v3

from micado import exceptions


class Authenticator:
    auth_url = None

    def authenticate(self):
        """ Returns a v3 auth object """
        raise NotImplementedError


class PasswordAuthenticator(Authenticator):
    def __init__(self, *, username, password):
        self.username = username
        self.password = password
        self.user_domain_name = "default"
        self.project_id = None

    def authenticate(self):
        """ Returns a Password auth object"""
        return v3.Password(
            self.auth_url,
            username=self.username,
            password=self.password,
            user_domain_name=self.user_domain_name,
            project_id=self.project_id,
        )


class OidcAuthenticator(Authenticator):
    def __init__(self, *, access_token, identity_provider, protocol="openid"):
        self.access_token = _verify_access_token(access_token)
        self.identity_provider = identity_provider
        self.protocol = protocol
        self.project_id = None

    def authenticate(self):
        """ Returns an OidcAccessToken auth object"""
        return v3.OidcAccessToken(
            self.auth_url,
            protocol=self.protocol,
            access_token=self.access_token,
            identity_provider=self.identity_provider,
            project_id=self.project_id,
        )


def _verify_access_token(access_token):
    if isinstance(access_token, dict):
        return refresh_openid_token(**access_token)
    return access_token


class AppCredAuthenticator(Authenticator):
    def __init__(
        self, *, application_credential_id, application_credential_secret
    ):
        self.application_credential_id = application_credential_id
        self.application_credential_secret = application_credential_secret

    def authenticate(self):
        """ Returns an ApplicationCredential auth object"""
        return v3.ApplicationCredential(
            self.auth_url,
            application_credential_id=self.application_credential_id,
            application_credential_secret=self.application_credential_secret,
        )


def refresh_openid_token(
    *,
    url,
    refresh_token,
    client_id="token-portal",
    client_secret=None,
    grant_type="refresh_token",
    scope="openid email profile",
):
    """Returns a new OpenID access token"""
    body = {
        "client_id": client_id,
        "grant_type": grant_type,
        "refresh_token": refresh_token,
        "scope": scope,
    }
    post_req = {"url": url, "data": body}

    if client_secret:
        body["client_secret"] = client_secret
        post_req["auth"] = (client_id, client_secret)

    try:
        response = requests.post(**post_req).json()
        return response["access_token"]
    except KeyError:
        raise exceptions.MicadoException("OpenID failed: {response}")


AUTH_TYPES = (
    PasswordAuthenticator,
    AppCredAuthenticator,
    OidcAuthenticator,
)



import logging
import logging.config
import os
import string
import secrets
from pathlib import Path

from Crypto.PublicKey import RSA
from ruamel.yaml import YAML

DEFAULT_PATH = Path.home() / ".micado-cli"

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
micado_cli_dir = Path(os.environ.get("MICADO_CLI_DIR", DEFAULT_PATH))
micado_cli_dir.mkdir(parents=True, exist_ok=True)
ch = logging.StreamHandler()
fh = logging.handlers.RotatingFileHandler(
    filename=str(micado_cli_dir / "micado-cli.log"),
    mode="a",
    maxBytes=52428800,
    backupCount=3,
)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s : %(message)s")
ch.setFormatter(formatter)
fh.setFormatter(formatter)
logger.addHandler(ch)
logger.addHandler(fh)


class DataHandling:
    @staticmethod
    def persist_data(path, server_id, **kwargs):
        """Persist data to in a file

        Args:
            path (string): Persist file location
            server_id (string): MiCADO UUID
        """
        content = None
        data = list()
        args = dict()
        for key, value in kwargs.items():
            args[key] = value
        data.append({server_id: args})
        yaml = YAML()
        yaml.indent(mapping=2, sequence=4, offset=2)
        if os.path.isfile(path):
            logger.debug("Data file exist...")
            with open(path) as f:
                content = yaml.load(f)
            # Handle edge cases when the file is empty or it has syntax error
            try:
                content["micados"] += data
            except:
                content = dict()
                content["micados"] = data
        else:
            logger.debug("Data file does not exist. Creating new file...")
            content = {"micados": data}
        with open(path, "w") as f:
            yaml.dump(content, f)

    @staticmethod
    def get_properties(path, server_id):
        """Return properties from a server_id

        Args:
            path (string): File location
            server_id (string): MiCADO UUID

        Raises:
            Exception: return with exception when the uuid can't find in the file

        Returns:
            (string): server properties
        """
        yaml = YAML()
        content = None
        try:
            with open(path, mode="r") as f:
                content = yaml.load(f)
        except Exception as e:
            raise e
        search = [i for i in content["micados"] if i.get(server_id, None)]
        if not search:
            logger.error("Can't find {} record!".format(server_id))
            raise LookupError("Can't find property!")
        else:
            return search[0][server_id]

    @staticmethod
    def update_data(path, server_id, **kwargs):
        """Updata UUID properties in the file.

        Args:
            path (string): File location
            server_id (string): MiCADO UUID
        """
        content = None
        args = dict()
        for key, value in kwargs.items():
            args[key] = value
        args = {server_id: args}
        yaml = YAML()
        yaml.indent(mapping=2, sequence=4, offset=2)
        if os.path.isfile(path):
            logger.debug("Data file exist...")
            with open(path) as f:
                content = yaml.load(f)
            server = DataHandling.get_properties(path, server_id)
            args[server_id].update(server)
            for i in content["micados"]:
                if i.get(server_id, None) != None:
                    i.update(args)
                    logger.debug("Data updated...")
                    break
        with open(path, "w") as f:
            yaml.dump(content, f)


class SSHKeyHandling:
    @staticmethod
    def get_pub_key(home):
        """Get public config key from home location.

        Returns:
            string: Public config key
        """
        if not SSHKeyHandling._check_ssh_key_existance(home):
            SSHKeyHandling._create_ssh_keys(home)
        with open(home + "micado_cli_config_pub_key", "r") as f:
            pub_key = f.readline()
        return pub_key

    @staticmethod
    def _check_ssh_key_existance(home):
        """Check if SSH config key exist.

        Returns:
            boolean: True if it is exist
        """
        return os.path.isfile(home + "micado_cli_config_priv_key") and os.path.isfile(
            home + "micado_cli_config_pub_key"
        )

    @staticmethod
    def _create_ssh_keys(home):
        """Create SSH config key, and set the correct permission."""
        key = RSA.generate(2048)
        with open(home + "micado_cli_config_priv_key", "wb") as f:
            f.write(key.export_key(format="PEM"))
        with open(home + "micado_cli_config_pub_key", "wb") as f:
            f.write(key.public_key().export_key(format="OpenSSH"))
        os.chmod(home + "micado_cli_config_priv_key", 0o600)
        os.chmod(home + "micado_cli_config_pub_key", 0o666)

def generate_password():
    alphabet = string.ascii_letters + string.digits
    password = "".join(secrets.choice(alphabet) for i in range(14))
    return password



"""

Base class for models and resource collections in MiCADO

"""


class Model:
    """Generic class for models of objects in MiCADO

    """

    def __init__(self, client, id=None, info=None, resource=None):
        self.client = client
        self.id = id
        self.info = info or {}
        self.resource = resource

    def reload(self):
        if not self.resource:
            pass
        updated = self.resource.get(self.id)
        self.info = updated.info


class Resource:
    """Generic class for collections of resources in MiCADO

    """

    model = None

    def __init__(self, client=None):
        self.client = client

    def get(self):
        raise NotImplementedError

    def list(self):
        raise NotImplementedError

    def create(self):
        raise NotImplementedError

    def _make_model(self, id, info):
        return self.model(self.client, id, info, self)
